@@ -1,4 +1,8 @@
+"""
+Copyright 2020 The Microsoft DeepSpeed Team
+"""
 import os
+import sys
 import time
 import torch
 import importlib
@@ -44,14 +48,31 @@ def get_default_compute_capatabilities():
     return compute_caps
 
 
+# list compatible minor CUDA versions - so that for example pytorch built with cuda-11.0 can be used
+# to build deepspeed and system-wide installed cuda 11.2
+cuda_minor_mismatch_ok = {
+    10: ["10.0",
+         "10.1",
+         "10.2"],
+    11: ["11.0",
+         "11.1",
+         "11.2",
+         "11.3"],
+}
+
+
 def assert_no_cuda_mismatch():
     cuda_major, cuda_minor = installed_cuda_version()
     sys_cuda_version = f'{cuda_major}.{cuda_minor}'
     torch_cuda_version = ".".join(torch.version.cuda.split('.')[:2])
     # This is a show-stopping error, should probably not proceed past this
     if sys_cuda_version != torch_cuda_version:
-        if sys_cuda_version == "11.1" and torch_cuda_version == "11.0":
-            # it works to build against installed cuda-11.1 while torch was built with cuda-11.0
+        if (cuda_major in cuda_minor_mismatch_ok
+                and sys_cuda_version in cuda_minor_mismatch_ok[cuda_major]
+                and torch_cuda_version in cuda_minor_mismatch_ok[cuda_major]):
+            print(f"Installed CUDA version {sys_cuda_version} does not match the "
+                  f"version torch was compiled with {torch.version.cuda} "
+                  "but since the APIs are compatible, accepting this combination")
             return
         raise Exception(
             f"Installed CUDA version {sys_cuda_version} does not match the "
@@ -119,6 +140,37 @@ class OpBuilder(ABC):
         '''
         return True
 
+    def extra_ldflags(self):
+        return []
+
+    def libraries_installed(self, libraries):
+        valid = False
+        check_cmd = 'dpkg -l'
+        for lib in libraries:
+            result = subprocess.Popen(f'dpkg -l {lib}',
+                                      stdout=subprocess.PIPE,
+                                      stderr=subprocess.PIPE,
+                                      shell=True)
+            valid = valid or result.wait() == 0
+        return valid
+
+    def simd_width(self):
+        if not self.command_exists('lscpu'):
+            self.warning(
+                f"{self.name} is attempted to query 'lscpu' to detect the existence "
+                "of AVX instructions. However, 'lscpu' does not appear to exist on "
+                "your system, will fall back to non-vectorized execution.")
+            return ''
+
+        result = subprocess.check_output('lscpu', shell=True)
+        result = result.decode('utf-8').strip().lower()
+        if 'genuineintel' in result:
+            if 'avx512' in result:
+                return '-D__AVX512__'
+            elif 'avx2' in result:
+                return '-D__AVX256__'
+        return ''
+
     def python_requirements(self):
         '''
         Override if op wants to define special dependencies, otherwise will
@@ -165,7 +217,8 @@ class OpBuilder(ABC):
         return CppExtension(name=self.absolute_name(),
                             sources=self.sources(),
                             include_dirs=self.include_paths(),
-                            extra_compile_args={'cxx': self.cxx_args()})
+                            extra_compile_args={'cxx': self.cxx_args()},
+                            extra_link_args=self.extra_ldflags())
 
     def load(self, verbose=True):
         from ...git_version_info import installed_ops, torch_info
@@ -213,6 +266,7 @@ class OpBuilder(ABC):
             ],
             extra_cflags=self.cxx_args(),
             extra_cuda_cflags=self.nvcc_args(),
+            extra_ldflags=self.extra_ldflags(),
             verbose=verbose)
         build_duration = time.time() - start_build
         if verbose:
@@ -298,7 +352,20 @@ class CUDAOpBuilder(OpBuilder):
         return CUDAExtension(name=self.absolute_name(),
                              sources=self.sources(),
                              include_dirs=self.include_paths(),
+                             libraries=self.libraries_args(),
                              extra_compile_args={
                                  'cxx': self.cxx_args(),
                                  'nvcc': self.nvcc_args()
                              })
+
+    def cxx_args(self):
+        if sys.platform == "win32":
+            return ['-O2']
+        else:
+            return ['-O3', '-std=c++14', '-g', '-Wno-reorder']
+
+    def libraries_args(self):
+        if sys.platform == "win32":
+            return ['cublas', 'curand']
+        else:
+            return []
